@@ -8,6 +8,10 @@ lookback <- 1:15
 data_train <- create_lagged_df(data_seatbelts, type = "train", outcome_col = 1,
                                lookback = lookback, horizon = horizons)
 
+# All historical window lengths of 12 plus any partial windows at the end of the dataset.
 windows <- create_windows(data_train, window_length = 12)
+plot(windows, data_train)
 
+# Two custom validation windows with different lengths.
+windows <- create_windows(data_train, window_start = c(20, 80), window_stop = c(30, 100))
 plot(windows, data_train)
