@@ -69,6 +69,7 @@ test_that("return_error produces correct mae and mape across aggregations", {
   )
 })
 #------------------------------------------------------------------------------
+#------------------------------------------------------------------------------
 
 test_that("validation error, window_length = 0 and 1 horizon returns the same error metrics across all aggregations", {
 
@@ -124,6 +125,7 @@ test_that("validation error, window_length = 0 and 1 horizon returns the same er
   expect_mapequal(data_error[[2]], data_error[[3]])
 })
 #------------------------------------------------------------------------------
+#------------------------------------------------------------------------------
 
 test_that("test/forecast error, 1 horizon returns the 2 data.frames of error metrics", {
 
@@ -175,3 +177,51 @@ test_that("test/forecast error, 1 horizon returns the 2 data.frames of error met
     nrow(data_error[[3]]) == 1
       )
 })
+#------------------------------------------------------------------------------
+#------------------------------------------------------------------------------
+
+test_that("return_error returns a data.frame when the input is 1 model from combine_forecasts(type = 'horizon')", {
+
+  data("data_seatbelts", package = "forecastML")
+
+  horizons <- c(1, 12)
+  lookback <- 1:15
+
+  data_train <- create_lagged_df(data_seatbelts, type = "train", outcome_col = 1,
+                                 lookback = lookback, horizon = horizons)
+
+  windows <- create_windows(data_train, window_length = 0)
+
+  model_function <- function(data, my_outcome_col) {
+    model <- lm(DriversKilled ~ ., data = data)
+    return(model)
+  }
+
+  set.seed(224)
+  model_results <- train_model(data_train, windows, model_name = "LASSO", model_function)
+
+  data_forecast <- create_lagged_df(data_seatbelts, type = "forecast", outcome_col = 1,
+                                    lookback = lookback, horizon = horizons)
+
+  prediction_function <- function(model, data_features) {
+
+    x <- data_features
+
+    data_pred <- data.frame("y_pred" = predict(model, newdata = x))
+    return(data_pred)
+  }
+
+  data_forecasts <- predict(model_results, prediction_function = list(prediction_function), data = data_forecast)
+
+  data_combined <- combine_forecasts(data_forecasts)
+
+  set.seed(224)
+  data_test <- data.frame("DriversKilled" = runif(12, 110, 160), "index" = 193:(193 + 11))
+  test_indices <- data_test$index
+
+  data_error_combined <- return_error(data_combined, data_test, test_indices)
+
+  all(
+    methods::is(data_error_combined, "data.frame")
+  )
+})
